"""Escea Fireplace UDP messaging module

   Implements simple UDP messages to Fireplace and receiving responses
"""

import asyncio
import logging

from asyncio import Lock
from asyncio.base_events import BaseEventLoop
from async_timeout import timeout
from typing import Any, Dict

# Pescea imports:
from .message import Message, CommandID, expected_response
from .udp_endpoints import open_local_endpoint, open_remote_endpoint


_LOG = logging.getLogger(__name__)

# Port used for discovery and integration
# (same port is used for replies)
CONTROLLER_PORT = 3300

# Time to wait for results from server
REQUEST_TIMEOUT = 5.0

Responses = Dict[str, Message]


class Datagram:
    """Send UDP Datagrams to fireplace and receive responses"""

    def __init__(
        self, event_loop: BaseEventLoop, device_ip: str, sending_lock: Lock
    ) -> None:
        """Create a simple datagram client interface.

        Args:
            event_loop: loop to use for coroutines
            device_addr: Device network address. Usually specified as IP
                address (can be a broadcast address in the case of fireplace search)
            sending_lock: Provided to attempt to make thread safe

        Raises:
            ConnectionRefusedError: If no Escea fireplace is discovered, or no
                device discovered at the given IP address, or the UID does not match
        """
        self._ip = device_ip
        self._event_loop = event_loop
        self.sending_lock = sending_lock

    @property
    def ip(self) -> str:
        """Target IP address"""
        return self._ip

    def set_ip(self, ip_addr: str) -> None:
        """Change the Target IP address"""
        self._ip = ip_addr

    async def send_command(self, command: CommandID, data: Any = None) -> Responses:
        """Send command via UDP
        Returns received response(s) and IP addresses they come from

        Args:
        - command: Fireplace command (refer Message)
        - data: ignored except for setting desired temperature

        Raises ConnectionError if unable to send command
        """
        message = Message(command=command, set_temp=data)
        responses = dict()  # type: Responses
        broadcast = command == CommandID.SEARCH_FOR_FIRES
        local = None
        remote = None

        # set up receiver before we send anything
        async with self.sending_lock:
            try:
                local = await open_local_endpoint(
                    port=CONTROLLER_PORT,
                    loop=self._event_loop,
                    reuse_port=True,
                )
                remote = await open_remote_endpoint(
                    host=self._ip,
                    port=CONTROLLER_PORT,
                    loop=self._event_loop,
                    allow_broadcast=broadcast,
                )
                remote.send(message.bytearray_)
                remote.close()
                async with timeout(REQUEST_TIMEOUT):
                    while True:

                        data, (addr, _) = await local.receive()
                        response = Message(incoming=data)
                        if response.is_command:
                            if not broadcast:
                                _LOG.error(
                                    "Unexpected command id: %s", response.command_id
                                )
                        else:  # response
                            if response.response_id != expected_response(command):
                                _LOG.debug(
                                    "Message response id: %s does not match command id: %s",
                                    response.response_id,
                                    command,
                                )
                            else:
                                responses[addr] = response
                        if not broadcast:
                            break
                    local.close()
            except (asyncio.TimeoutError, ValueError):
                pass
            finally:
                if remote is not None and not remote.closed:
                    remote.close()
                if local is not None and not local.closed:
                    local.close()

        if len(responses) == 0:
            _LOG.debug(
                "Unable to send UDP message - Local endpoint closed:%s, Remote endpoint closed:%s",
                "None" if local is None else local.closed,
                "None" if remote is None else remote.closed,
            )
            raise ConnectionError("Unable to send/receive UDP message")

        return responses



""" Interface to the Escea fireplace controller

    Interaction through the Controller class.
    Discovery provides search for fireplaces and support for listeners.

    The primary purpose of this is to be integrated into Home Assistant,
    however it will work with other integration systems that support python.
"""

from .controller import Controller
from .discovery import Listener, AbstractDiscoveryService, discovery_service

__ALL__ = [Controller, AbstractDiscoveryService, Listener, discovery_service]



"""Escea Fireplace Message module"""

import logging

from enum import Enum

_LOG = logging.getLogger(__name__)

# The same message structure is used for commands and responses:
MESSAGE_LENGTH = 15
MSG_OFFSET_START_BYTE = 0  # Byte 1: Start Byte code
MSG_OFFSET_ID = 1  # Byte 2: Command / Response ID
MSG_OFFSET_DATA_LENGTH = 2  # Byte 3: Data Length

# Byte 4..13: Data (0 filled / don'tcare after Data Length)
MSG_OFFSET_DATA_START = 3
MSG_OFFSET_DATA_END = 12

# Byte 14: CRC (Sum Bytes 2 to 13, overflowing on 256)
MSG_OFFSET_CRC = 13
MSG_OFFSET_END_BYTE = 14  # Byte 15: End Byte code

# Data structure for STATUS response:
# Data Byte 1: (boolean) Fireplace has new timers (not used)
DATA_OFFSET_TIMERS = 0
DATA_OFFSET_FIRE_ON = 1  # Data Byte 2: (boolean) Fire is On
DATA_OFFSET_BOOST_ON = 2  # Data Byte 3: (boolean) Fan Boost is on
DATA_OFFSET_EFFECT_ON = 3  # Data Byte 4: (boolean) Flame Effect is on

# Data Byte 5: (unsigned int) Desired Temperature
DATA_OFFSET_DESIRED_TEMP = 4

# Data Byte 6: (unsigned int) Room Temperature
DATA_OFFSET_CURRENT_TEMP = 5

# Data structure for I_AM_A_FIRE response:
# Data Bytes 1..4: (Unsigned Long, big Endian) Serial Number (use for UID)
DATA_OFFSET_SERIAL = 0
# Data Byte 5..6: (Unsigned Long, big Endian) PIN (not used)
DATA_OFFSET_PIN = 4

# Preconfigured start/end characters:
MESSAGE_START_BYTE = 0x47
MESSAGE_END_BYTE = 0x46

# Valid command identifiers:
class CommandID(Enum):
    STATUS_PLEASE = 0x31
    POWER_ON = 0x39
    POWER_OFF = 0x3A
    SEARCH_FOR_FIRES = 0x50
    FAN_BOOST_ON = 0x37
    FAN_BOOST_OFF = 0x38
    FLAME_EFFECT_ON = 0x56
    FLAME_EFFECT_OFF = 0x55
    NEW_SET_TEMP = 0x57


class ResponseID(Enum):
    STATUS = 0x80
    POWER_ON_ACK = 0x8D
    POWER_OFF_ACK = 0x8F
    FAN_BOOST_ON_ACK = 0x89
    FAN_BOOST_OFF_ACK = 0x8B
    FLAME_EFFECT_ON_ACK = 0x61
    FLAME_EFFECT_OFF_ACK = 0x60
    NEW_SET_TEMP_ACK = 0x66
    I_AM_A_FIRE = 0x90


# Acceptable limits when commanding NEW_SET_TEMP:
MIN_SET_TEMP = 4
MAX_SET_TEMP = 30


def expected_response(command: CommandID) -> ResponseID:
    """Utility function to check correct response
    Raises ValueError (if unexpected CommandID)
    """
    if command == CommandID.STATUS_PLEASE:
        return ResponseID.STATUS
    elif command == CommandID.POWER_ON:
        return ResponseID.POWER_ON_ACK
    elif command == CommandID.POWER_OFF:
        return ResponseID.POWER_OFF_ACK
    elif command == CommandID.SEARCH_FOR_FIRES:
        return ResponseID.I_AM_A_FIRE
    elif command == CommandID.FAN_BOOST_ON:
        return ResponseID.FAN_BOOST_ON_ACK
    elif command == CommandID.FAN_BOOST_OFF:
        return ResponseID.FAN_BOOST_OFF_ACK
    elif command == CommandID.FLAME_EFFECT_ON:
        return ResponseID.FLAME_EFFECT_ON_ACK
    elif command == CommandID.FLAME_EFFECT_OFF:
        return ResponseID.FLAME_EFFECT_OFF_ACK
    elif command == CommandID.NEW_SET_TEMP:
        return ResponseID.NEW_SET_TEMP_ACK
    else:
        raise (ValueError)


class Message:

    """Implements messages to and from the fireplace.
    Refer to Escea Fireplace LAN Comms Spec for details.
    """

    def _initialise_data(self) -> None:
        """Default all attributes"""
        self._is_command = None
        self._id = None
        self._has_new_timers = None
        self._fan_boost_on = None
        self._effect_on = None
        self._desired_temp = None
        self._current_temp = None
        self._serial = None
        self._pin = None
        self._bytearray = None
        self._crc_sum = None

    def __init__(
        self,
        command: CommandID = None,
        set_temp: int = None,
        incoming: bytearray = None,
    ) -> None:
        """Creates either a Command or a Response message

        Args: Either command or incoming must be set
        """

        if command is not None:
            self._create_command(command, set_temp)

        elif incoming is not None:
            self._parse_incoming(incoming)

        else:
            raise (ValueError, "Invalid constructor")

    def _create_command(self, command: CommandID, set_temp: float = None):
        """Create a command (outgoing) message.
        Args:
        - command: valid command code to fireplace
        - set_temp: desired temperature (only applies to that command)

        Use the property bytearray_of to get the bytes to send.
        """
        self._initialise_data()

        self._is_command = True
        self._id = command
        self._desired_temp = set_temp  # Only used for NEW_SET_TEMP

        # Build outgoing message:

        self._bytearray = bytearray(MESSAGE_LENGTH)

        self._bytearray[MSG_OFFSET_START_BYTE] = MESSAGE_START_BYTE
        self._bytearray[MSG_OFFSET_END_BYTE] = MESSAGE_END_BYTE

        self._bytearray[MSG_OFFSET_ID] = (self._id).value

        if self._id == CommandID.NEW_SET_TEMP:
            if set_temp is None or set_temp < MIN_SET_TEMP or set_temp > MAX_SET_TEMP:
                raise ValueError
            self._bytearray[MSG_OFFSET_DATA_LENGTH] = 1
            self._bytearray[MSG_OFFSET_DATA_START] = set_temp

        # Calculate CRC
        self._crc_sum = 0
        for i in range(MSG_OFFSET_ID, MSG_OFFSET_DATA_END):
            self._crc_sum += self._bytearray[i]
        self._crc_sum = self._crc_sum % 256
        self._bytearray[MSG_OFFSET_CRC] = self._crc_sum

    def _parse_incoming(self, incoming: bytearray):
        """Create a response Message from incoming buffer

        Note:
            Can also be used with a Command bytearray for test purposes

        Raises:
            ValueError if message content does not match specification
        """
        self._initialise_data()

        self._bytearray = incoming

        # Check message integrity
        if len(incoming) != MESSAGE_LENGTH:
            raise ValueError(
                "Message: '{}' has incorrect message length: {} (expecting {})".format(
                    incoming.hex(), len(incoming), MESSAGE_LENGTH
                )
            )

        if incoming[MSG_OFFSET_START_BYTE] != MESSAGE_START_BYTE:
            raise ValueError(
                "Message: '{}' has invalid start byte: {} (expecting {})".format(
                    incoming.hex(), incoming[MSG_OFFSET_START_BYTE], MESSAGE_START_BYTE
                )
            )

        if incoming[MSG_OFFSET_END_BYTE] != MESSAGE_END_BYTE:
            raise ValueError(
                "Message: '{}' has invalid end byte: {} (expecting {})".format(
                    incoming.hex(), incoming[MSG_OFFSET_END_BYTE], MESSAGE_END_BYTE
                )
            )

        # Check CRC
        self._crc_sum = 0
        for i in range(MSG_OFFSET_ID, MSG_OFFSET_DATA_END):
            self._crc_sum += incoming[i]

        self._crc_sum = self._crc_sum % 256
        if self._crc_sum != incoming[MSG_OFFSET_CRC]:
            raise ValueError(
                "Message: '{}' has invalid CRC: {} (expecting {})".format(
                    incoming.hex(), incoming[MSG_OFFSET_CRC], self._crc_sum
                )
            )

        id_value = incoming[MSG_OFFSET_ID]
        if id_value in ResponseID._value2member_map_:

            # Normal use: Decode incoming responses from fireplace
            self._parse_response(incoming)

        elif incoming[MSG_OFFSET_ID] in CommandID._value2member_map_:
            # For Test use only: Decode incoming command bytearray
            self._parse_command(incoming)

        else:
            raise ValueError("Invalid message id: {}".format(id_value))

    def _parse_response(self, incoming: bytearray):
        """Normal use case - decode incoming response from fireplace"""

        self._is_command = False
        self._id = ResponseID(incoming[MSG_OFFSET_ID])

        # Extract data
        if (self._id) == ResponseID.STATUS:
            if incoming[MSG_OFFSET_DATA_LENGTH] != 6:
                raise ValueError(
                    "Message: '{}' has invalid data length: {} (expecting 6)".format(
                        incoming.hex(), incoming[MSG_OFFSET_DATA_LENGTH]
                    )
                )

            self._has_new_timers = bool(
                incoming[MSG_OFFSET_DATA_START + DATA_OFFSET_TIMERS]
            )
            self._fire_on = bool(incoming[MSG_OFFSET_DATA_START + DATA_OFFSET_FIRE_ON])
            self._fan_boost_on = bool(
                incoming[MSG_OFFSET_DATA_START + DATA_OFFSET_BOOST_ON]
            )
            self._effect_on = bool(
                incoming[MSG_OFFSET_DATA_START + DATA_OFFSET_EFFECT_ON]
            )
            self._desired_temp = int(
                incoming[MSG_OFFSET_DATA_START + DATA_OFFSET_DESIRED_TEMP]
            )
            self._current_temp = int(
                incoming[MSG_OFFSET_DATA_START + DATA_OFFSET_CURRENT_TEMP]
            )

        elif (self._id) == ResponseID.I_AM_A_FIRE:
            if incoming[MSG_OFFSET_DATA_LENGTH] != 6:
                # Just log this... there is an error on the fireplace side here
                _LOG.debug(
                    "Message: %s Has Invalid Data Length: %s (expecting 6)",
                    incoming.hex(),
                    str(incoming[MSG_OFFSET_DATA_LENGTH]),
                )
            self._serial = int.from_bytes(
                incoming[
                    MSG_OFFSET_DATA_START
                    + DATA_OFFSET_SERIAL : MSG_OFFSET_DATA_START
                    + DATA_OFFSET_SERIAL
                    + 4
                ],
                byteorder="big",
                signed=False,
            )
            self._pin = int.from_bytes(
                incoming[
                    MSG_OFFSET_DATA_START
                    + DATA_OFFSET_PIN : MSG_OFFSET_DATA_START
                    + DATA_OFFSET_PIN
                    + 2
                ],
                byteorder="big",
                signed=False,
            )

        else:
            if int(incoming[MSG_OFFSET_DATA_LENGTH]) != 0:
                raise ValueError(
                    "Message: '{}' has invalid data length: {} (expecting 0)".format(
                        incoming.hex(), int(incoming[MSG_OFFSET_DATA_LENGTH])
                    )
                )

    def _parse_command(self, incoming: bytearray):
        """For Test use only: Decode incoming command bytearray"""
        self._is_command = True
        self._id = CommandID(incoming[MSG_OFFSET_ID])

        if self._id == CommandID.NEW_SET_TEMP:
            self._desired_temp = incoming[MSG_OFFSET_DATA_START]

        if incoming[MSG_OFFSET_DATA_LENGTH] != (
            1 if self._id == CommandID.NEW_SET_TEMP else 0
        ):
            raise ValueError(
                "Message: '{}' has invalid data length: {} (expecting 1)".format(
                    incoming.hex(), incoming[MSG_OFFSET_DATA_LENGTH]
                )
            )

    @property
    def is_command(self) -> bool:
        return self._is_command

    @property
    def is_response(self) -> bool:
        return not self._is_command

    @property
    def command_id(self) -> CommandID:
        if self._is_command:
            return self._id
        else:
            raise (ValueError)

    @property
    def response_id(self) -> ResponseID:
        if not self._is_command:
            return self._id
        else:
            raise (ValueError)

    @property
    def has_new_timers(self) -> bool:
        return self._has_new_timers

    @property
    def fire_is_on(self) -> bool:
        return self._fire_on

    @property
    def fan_boost_is_on(self) -> bool:
        return self._fan_boost_on

    @property
    def flame_effect(self) -> bool:
        return self._effect_on

    @property
    def desired_temp(self) -> int:
        return self._desired_temp

    @property
    def current_temp(self) -> int:
        return self._current_temp

    @property
    def serial_number(self) -> int:
        return self._serial

    @property
    def pin(self) -> int:
        return self._pin

    @property
    def crc(self) -> int:
        return self._crc_sum

    @property
    def bytearray_(self) -> bytearray:
        return self._bytearray

    # Internal use test methods follow:

    def mock_response(
        response_id: ResponseID,
        uid: int = 0,
        has_new_timers: bool = False,
        fire_on: bool = False,
        fan_boost_on: bool = False,
        effect_on: bool = False,
        desired_temp: int = MAX_SET_TEMP,
        current_temp: int = MIN_SET_TEMP,
        force_crc_error: bool = False,
        force_id_error: bool = False,
        force_data_len_error: bool = False,
        force_start_byte_error: bool = False,
        force_end_byte_error: bool = False,
    ) -> bytearray:
        """Create a dummy message for testing purposes."""

        message = bytearray(MESSAGE_LENGTH)

        if force_start_byte_error:
            message[MSG_OFFSET_START_BYTE] = 0
        else:
            message[MSG_OFFSET_START_BYTE] = MESSAGE_START_BYTE

        if force_end_byte_error:
            message[MSG_OFFSET_END_BYTE] = 0
        else:
            message[MSG_OFFSET_END_BYTE] = MESSAGE_END_BYTE

        if force_id_error:
            message[MSG_OFFSET_ID] = 0
        else:
            message[MSG_OFFSET_ID] = response_id.value

        if response_id == ResponseID.STATUS:
            message[MSG_OFFSET_DATA_LENGTH] = 6
            message[MSG_OFFSET_DATA_START] = desired_temp
            message[MSG_OFFSET_DATA_START + DATA_OFFSET_TIMERS] = has_new_timers
            message[MSG_OFFSET_DATA_START + DATA_OFFSET_FIRE_ON] = fire_on
            message[MSG_OFFSET_DATA_START + DATA_OFFSET_BOOST_ON] = fan_boost_on
            message[MSG_OFFSET_DATA_START + DATA_OFFSET_EFFECT_ON] = effect_on
            message[MSG_OFFSET_DATA_START + DATA_OFFSET_DESIRED_TEMP] = desired_temp
            message[MSG_OFFSET_DATA_START + DATA_OFFSET_CURRENT_TEMP] = current_temp

        elif response_id == ResponseID.I_AM_A_FIRE:
            message[MSG_OFFSET_DATA_LENGTH] = 6
            serial_segment = uid.to_bytes(length=4, byteorder="big", signed=False)
            pin_segment = int(9999).to_bytes(length=2, byteorder="big", signed=False)
            for i in range(len(serial_segment)):
                message[
                    MSG_OFFSET_DATA_START + DATA_OFFSET_SERIAL + i
                ] = serial_segment[i]
            for i in range(len(pin_segment)):
                message[MSG_OFFSET_DATA_START + DATA_OFFSET_PIN + i] = pin_segment[i]

        else:
            message[MSG_OFFSET_DATA_LENGTH] = 0

        if force_data_len_error:
            message[MSG_OFFSET_DATA_LENGTH] += 1

        if force_crc_error:
            message[MSG_OFFSET_CRC] = 0
        else:
            # Calculate CRC
            crc_sum = 0
            for i in range(MSG_OFFSET_ID, MSG_OFFSET_DATA_END):
                crc_sum += message[i]
            crc_sum = crc_sum % 256
            message[MSG_OFFSET_CRC] = crc_sum

        return message



"""Escea Network Controller module"""

import logging
import asyncio

from enum import Enum
from typing import Dict, Union
from time import time
from async_timeout import timeout
from copy import deepcopy

# Pescea imports:
from .message import (
    Message,
    CommandID,
    ResponseID,
    MIN_SET_TEMP,
    MAX_SET_TEMP,
    expected_response,
)
from .datagram import Datagram

_LOG = logging.getLogger(__name__)

# Seconds between (internal) updates under normal conditions
REFRESH_INTERVAL = 30.0

# Seconds between publishing changes to listeners
# - Updates due to changes happen immediately
NOTIFY_REFRESH_INTERVAL = 5 * 60.0

# Retry rate when first stop getting responses
# - this is needed as UDP is not guaranteed and expect missed messages
RETRY_INTERVAL = 10.0

# Timeout to stop retrying and reduce poll rate
# - plus notify listeners of disconnections
RETRY_TIMEOUT = 60.0

# Time to retry updates once we declare disconnection
DISCONNECTED_INTERVAL = 5 * 60.0

# Time to wait for fireplace to turn fire on / off
# - While busy, commands are stored locally, but not sent on
# - This period was measured from an Escea OEM remote control
ON_OFF_BUSY_WAIT_TIME = 66.0


class Controller:
    """Interface to Escea controller"""

    class Fan(Enum):
        """Supported fan modes"""

        FLAME_EFFECT = "FlameEffect"
        AUTO = "Auto"
        FAN_BOOST = "FanBoost"

    class State(Enum):
        """Controller states:

        Under normal operations:
            The Controller is READY:
                - The Controller sends commands directly to the Fireplace
                - The Controller polls at REFRESH_INTERVAL
        When toggling the fire power:
            The Controller remains BUSY for ON_OFF_BUSY_WAIT_TIME:
                - The Controller buffers requests but does not send to the Fireplace
        When responses are missed (expected as we are using UDP datagrams):
            The Controller is NON_RESPONSIVE
                - The Controller will poll at a (quicker) retry rate
        When there are no comms for a prolonged period:
            The Controller enters DISCONNECTED state
                - The Controller will continue to poll at a reduced rate
                - The Controller buffers requests but cannot send to the Fireplace
        """

        BUSY = "BusyWaiting"
        READY = "Ready"
        NON_RESPONSIVE = "NonResponsive"
        DISCONNECTED = "Disconnected"

    class Settings(Enum):
        """Available controller system settings - Internal Use Only"""

        IP_ADDRESS = "IPAddress"
        DEVICE_UID = "DeviceUId"
        CONTROLLER_STATE = "State"
        HAS_NEW_TIMERS = "HasNewTimers"
        FIRE_IS_ON = "FireIsOn"
        FAN_MODE = "FanMode"
        DESIRED_TEMP = "DesiredTemp"
        CURRENT_TEMP = "CurrentTemp"

    Value = Union[str, int, float, bool, Fan]
    ControllerData = Dict[Settings, Value]

    def __init__(self, discovery, device_uid: str, device_ip: str) -> None:
        """Create a controller interface.

        Args:
            discovery: DiscoveryService() object implementing (at least):
                        - loop
                        - sending_lock
                        - controller_update (callback)
                        - controller_disconnected (callback)
                        - controller_reconnected (callback)
            device_uid: Controller UId as a string (Serial Number of unit)
            device_addr: Device network address. Usually specified as IP
                address
        """

        self._discovery = discovery
        self._system_settings = {}  # type: Controller.ControllerData
        self._prior_settings = {}  # type: Controller.ControllerData
        self._system_settings[Controller.Settings.IP_ADDRESS] = device_ip
        self._system_settings[Controller.Settings.DEVICE_UID] = device_uid

        self._datagram = Datagram(
            self._discovery.loop, device_ip, self._discovery.sending_lock
        )

        self._interrupt_poll_loop_sleep = asyncio.Condition()

        self._initialised = False

    async def initialize(self) -> None:
        """Initialize the controller, does not complete until the firplace has
        been contacted and current settings read.
        """

        self._state = Controller.State.READY
        self._last_response = 0.0  # Tracks last valid message received
        self._busy_end_time = 0.0  # Tracks when exit BUSY state
        self._last_update = 0.0  # To 'rate limit' the notifications to discovery
        self._closed = False  # To exit poll_loop when done

        # Read current state of fireplace
        await self._refresh_system(notify=False)

        self._initialised = True

        # Start regular polling for status updates
        self._poll_loop_task = self._discovery.loop.create_task(self._poll_loop())

    async def close(self):
        """Signal loop to exit, then wait till done"""
        if self._closed:
            return

        self._closed = True
        async with self._interrupt_poll_loop_sleep:
            self._interrupt_poll_loop_sleep.notify()
        await self._poll_loop_task

    async def _poll_loop(self) -> None:
        """Regularly poll for status update from fireplace.
        If Disconnected, retry based on how long ago we last had an update.
        """
        while not self._closed:

            try:
                await self._refresh_system()
            except Exception as exc:
                _LOG.exception("Unexpected exception (EXITING): %s", repr(exc))
                self._closed = True
                return

            _LOG.debug(
                "Polling unit %s at address %s (current state is %s)",
                self._system_settings[Controller.Settings.DEVICE_UID],
                self._system_settings[Controller.Settings.IP_ADDRESS],
                self._state,
            )

            if self._state == Controller.State.READY:
                sleep_time = REFRESH_INTERVAL
            elif self._state == Controller.State.NON_RESPONSIVE:
                sleep_time = RETRY_INTERVAL
            elif self._state == Controller.State.DISCONNECTED:
                sleep_time = DISCONNECTED_INTERVAL
            elif self._state == Controller.State.BUSY:
                sleep_time = max(self._busy_end_time - time(), 0.0)

            try:
                # Sleep for poll time, allow early wakeup
                async with timeout(sleep_time):
                    async with self._interrupt_poll_loop_sleep:
                        await self._interrupt_poll_loop_sleep.wait()
            except asyncio.TimeoutError:
                pass

    @property
    def device_ip(self) -> str:
        """IP Address of the unit"""
        return self._system_settings[Controller.Settings.IP_ADDRESS]

    @property
    def device_uid(self) -> str:
        """UId of the unit (serial number)"""
        return self._system_settings[Controller.Settings.DEVICE_UID]

    @property
    def discovery(self):
        """Handle to the discovery service"""
        return self._discovery

    @property
    def state(self) -> State:
        """Controller state"""
        return self._state

    @property
    def is_on(self) -> bool:
        """True if the fire is turned on"""
        return self._get_system_state(Controller.Settings.FIRE_IS_ON)

    async def set_on(self, value: bool) -> None:
        """Turn the fire on or off.
        Does not return until command has been sent to the fireplace.
        Note: After systems receives on or off command, must wait ON_OFF_BUSY_TIMEOUT
        before sending further commands (they will be buffered internally)
        """
        await self._set_system_state(Controller.Settings.FIRE_IS_ON, value)

    @property
    def fan(self) -> Fan:
        """The current fan level."""
        return self._get_system_state(Controller.Settings.FAN_MODE)

    async def set_fan(self, value: Fan) -> None:
        """The fan level.
        Does not return until command has been sent to the fireplace.

        Note: A peculiarity of this fireplace is that if we change the fan
        setting while off, it actually turns the fire on.
        """
        await self._set_system_state(Controller.Settings.FAN_MODE, value)

    @property
    def desired_temp(self) -> float:
        """fireplace DesiredTemp temperature."""
        return float(self._get_system_state(Controller.Settings.DESIRED_TEMP))

    async def set_desired_temp(self, value: float):
        """Fireplace DesiredTemp temperature.
        Does not return until command has been sent to the fireplace.

        This is the target temp for the fire
        Args:
            value: Valid settings are in range MIN_TEMP..MAX_TEMP
            at 1 degree increments (will be rounded)
        """
        degrees = round(value)
        if degrees < MIN_SET_TEMP or degrees > MAX_SET_TEMP:
            _LOG.error(
                "Desired Temp %s is out of range (%s-%s)",
                degrees,
                MIN_SET_TEMP,
                MAX_SET_TEMP,
            )
            return

        await self._set_system_state(Controller.Settings.DESIRED_TEMP, degrees)

    @property
    def current_temp(self) -> float:
        """The room air temperature"""
        return float(self._get_system_state(Controller.Settings.CURRENT_TEMP))

    @property
    def min_temp(self) -> float:
        """The minimum valid target (desired) temperature"""
        return float(MIN_SET_TEMP)

    @property
    def max_temp(self) -> float:
        """The maximum valid target (desired) temperature"""
        return float(MAX_SET_TEMP)

    async def _refresh_system(self, notify: bool = True) -> None:
        """Request fresh status from the fireplace.
        This is also where state changes are handled.

        Approach:

            if current state BUSY (and not timed out) -> return

            request status

            New status received:
                if prior state READY
                    update local system settings from received message
                else (prior state DISCONNECTED / NON_RESPONSIVE / BUSY (timeout))
                    sync buffered commands to fireplace
                    new state READY
                    if prior state DISCONNECTED:
                        notify discovery reconnected

            No status received:
                prior state *ANY*
                    if time since last response < RETRY_TIMEOUT
                        new state NON_RESPONSIVE
                    else
                        new state DISCONNECTED
                        notify discovery disconnected
        """
        if self._state != Controller.State.BUSY or time() >= self._busy_end_time:
            # Ok to fetch new status

            prior_state = self._state
            response = await self._request_status()
            if (response is not None) and (response.response_id == ResponseID.STATUS):
                # We have a valid response - the controller is communicating

                self._state = Controller.State.READY

                # These values are readonly, so copy them in any case
                self._system_settings[
                    Controller.Settings.HAS_NEW_TIMERS
                ] = response.has_new_timers
                self._system_settings[
                    Controller.Settings.CURRENT_TEMP
                ] = response.current_temp

                if prior_state == Controller.State.READY:

                    # Normal operation, update our internal values
                    self._system_settings[
                        Controller.Settings.DESIRED_TEMP
                    ] = response.desired_temp
                    if response.fan_boost_is_on:
                        self._system_settings[
                            Controller.Settings.FAN_MODE
                        ] = Controller.Fan.FAN_BOOST
                    elif response.flame_effect:
                        self._system_settings[
                            Controller.Settings.FAN_MODE
                        ] = Controller.Fan.FLAME_EFFECT
                    else:
                        self._system_settings[
                            Controller.Settings.FAN_MODE
                        ] = Controller.Fan.AUTO
                    self._system_settings[
                        Controller.Settings.FIRE_IS_ON
                    ] = response.fire_is_on

                else:

                    # We have come back to READY state.
                    # We need to try to sync buffered settings to fireplace

                    if (
                        response.desired_temp
                        != self._system_settings[Controller.Settings.DESIRED_TEMP]
                    ):
                        await self._set_system_state(
                            Controller.Settings.DESIRED_TEMP,
                            self._system_settings[Controller.Settings.DESIRED_TEMP],
                            sync=True,
                        )

                    if response.fan_boost_is_on:
                        response_fan = Controller.Fan.FAN_BOOST
                    elif response.flame_effect:
                        response_fan = Controller.Fan.FLAME_EFFECT
                    else:
                        response_fan = Controller.Fan.AUTO
                    if not self._system_settings[Controller.Settings.FIRE_IS_ON]:
                        # Escea controller has a quirk that turning on
                        # FAN_BOOST or FLAME_EFFECT actually turns on the fire
                        # ... so must avoid setting fan when 'synching' if want it to stay off
                        self._system_settings[
                            Controller.Settings.FAN_MODE
                        ] = response_fan
                    else:
                        if (
                            response_fan
                            != self._system_settings[Controller.Settings.FAN_MODE]
                        ):
                            await self._set_system_state(
                                Controller.Settings.FAN_MODE,
                                self._system_settings[Controller.Settings.FAN_MODE],
                                sync=True,
                            )

                    # Do power last, as then we go to BUSY state
                    if (
                        response.fire_is_on
                        != self._system_settings[Controller.Settings.FIRE_IS_ON]
                    ):
                        await self._set_system_state(
                            Controller.Settings.FIRE_IS_ON,
                            self._system_settings[Controller.Settings.FIRE_IS_ON],
                            sync=True,
                        )

                    if prior_state == Controller.State.DISCONNECTED:
                        self._discovery.controller_reconnected(self)

            else:
                # No / invalid response, need to check if we need to change state
                if time() - self._last_response < RETRY_TIMEOUT:
                    self._state = Controller.State.NON_RESPONSIVE
                else:
                    self._state = Controller.State.DISCONNECTED
                    if prior_state != Controller.State.DISCONNECTED:
                        self._discovery.controller_disconnected(self, TimeoutError)

        if notify and self._state != Controller.State.DISCONNECTED:
            # send an update to discovery if there have been any changes
            # or if haven't sent one in last NOTIFY_REFRESH_INTERVAL
            changes_found = False
            for entry in self._system_settings:
                if not entry in self._prior_settings or (
                    self._prior_settings[entry] != self._system_settings[entry]
                ):
                    changes_found = True
                    break
            if changes_found or (time() - self._last_update > NOTIFY_REFRESH_INTERVAL):
                self._last_update = time()
                self._prior_settings = deepcopy(self._system_settings)
                self._discovery.controller_update(self)

    async def _request_status(self) -> Message:
        """Send command to fireplace requesting current status"""
        try:
            responses = await self._datagram.send_command(CommandID.STATUS_PLEASE)
            if len(responses) > 0:
                this_response = next(iter(responses))  # only expecting one
                if responses[this_response].response_id == expected_response(
                    CommandID.STATUS_PLEASE
                ):
                    _LOG.debug(
                        "_request_status - send_command(success): %s",
                        str(self.device_uid),
                    )
                    self._last_response = time()
                    return responses[this_response]
        except ConnectionError:
            pass
        # If we get here... did not receive a response or not valid
        if self._state != Controller.State.DISCONNECTED:
            self._state = Controller.State.NON_RESPONSIVE
        _LOG.debug(
            "_request_status - send_command(failed): %s (now: %s)",
            str(self.device_uid),
            self._state,
        )
        return None

    def refresh_address(self, address):
        """Called from discovery to point controller to the IP address"""
        if self._system_settings[Controller.Settings.IP_ADDRESS] == address:
            return

        self._datagram.set_ip(address)
        self._system_settings[Controller.Settings.IP_ADDRESS] = address

        # signal the poll loop to wake up and ask for new status
        async def signal_loop(self):
            async with self._interrupt_poll_loop_sleep:
                self._interrupt_poll_loop_sleep.notify()

        self._discovery.loop.create_task(signal_loop(self))

    def _get_system_state(self, state: Settings):
        """Locally stored (buffered) value, or received from fireplace"""
        return self._system_settings[state]

    async def _set_system_state(self, state: Settings, value, sync: bool = False):
        """Send command to fireplace to change given state.
        Args:
            sync: state/value have been buffered -> send to fireplace
        """

        # nothing to do if not synching our state, and already have right state
        if (not sync) and (self._system_settings[state] == value):
            return

        _LOG.debug(
            "_set_system_state - uid: %s | %s from:%s to:%s  (sync:%s)",
            str(self.device_uid),
            str(state),
            str(self._system_settings[state]),
            str(value),
            str(sync),
        )

        # save the new value internally
        self._system_settings[state] = value

        # send it to the fireplace if asked to sync, or fireplace is not busy
        if sync or self._state == Controller.State.READY:

            command = None

            if state == Controller.Settings.FIRE_IS_ON:
                if value:
                    command = CommandID.POWER_ON
                else:
                    command = CommandID.POWER_OFF

            elif state == Controller.Settings.DESIRED_TEMP:
                command = CommandID.NEW_SET_TEMP

            elif state == Controller.Settings.FAN_MODE:

                # Fan is implemented via separate FLAME_EFFECT and FAN_BOOST commands
                # Any change to Fan will take one or two separate commands:
                #
                # To AUTO:
                # 1. turn off FAN_BOOST
                if value == Controller.Fan.AUTO:
                    command = CommandID.FAN_BOOST_OFF

                # To FAN_BOOST:
                # 1. Turn off FLAME_EFFECT
                elif value == Controller.Fan.FAN_BOOST:
                    command = CommandID.FLAME_EFFECT_OFF

                # To FLAME_EFFECT:
                # 1. Turn off FAN_BOOST
                elif value == Controller.Fan.FLAME_EFFECT:
                    command = CommandID.FAN_BOOST_OFF

            else:
                raise (AttributeError, "Unexpected state: {0}".format(state))

            if command is not None:
                valid_response = False
                try:
                    responses = await self._datagram.send_command(command, value)
                    if (len(responses) > 0) and (
                        responses[next(iter(responses))].response_id
                        == expected_response(command)
                    ):
                        valid_response = True
                        _LOG.debug(
                            "_set_system_state - send_command(success): %s -> %s",
                            str(self.device_uid),
                            str(command),
                        )

                except ConnectionError:
                    pass
                if valid_response:
                    self._last_response = time()
                else:
                    return

            if state == Controller.Settings.FAN_MODE:
                # Fan is implemented via separate FLAME_EFFECT and FAN_BOOST commands
                # Any change will take one or two separate commands:
                #
                # To AUTO:
                # 2. turn off FLAME_EFFECT
                if value == Controller.Fan.AUTO:
                    command = CommandID.FLAME_EFFECT_OFF

                # To FAN_BOOST:
                # 2. Turn on FAN_BOOST
                elif value == Controller.Fan.FAN_BOOST:
                    command = CommandID.FAN_BOOST_ON

                # To FLAME_EFFECT:
                # 2. Turn on FLAME_EFFECT
                else:
                    command = CommandID.FLAME_EFFECT_ON

                valid_response = False
                try:
                    responses = await self._datagram.send_command(command, value)
                    if (len(responses) > 0) and (
                        responses[next(iter(responses))].response_id
                        == expected_response(command)
                    ):
                        valid_response = True
                        _LOG.debug(
                            "_set_system_state - send_command(success): %s -> %s",
                            str(self.device_uid),
                            str(command),
                        )
                except ConnectionError:
                    pass
                if valid_response:
                    self._last_response = time()
                else:
                    return

        # Need to refresh immediately after setting
        # (unless synching, in which case the poll loop will update)
        if not sync:
            await self._refresh_system()

        # If get here, and just toggled the fireplace power... need to buffer for a while
        if state == Controller.Settings.FIRE_IS_ON:
            self._state = Controller.State.BUSY
            self._busy_end_time = time() + ON_OFF_BUSY_WAIT_TIME
